/* use as:
 * ./scrabble <dict> <board> <hand>|sort|sort -k9nr|less
 */
#include "scrabble.h"
#include "dict.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

int kelimelik_bot;

/* english letters */
letter english_letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 4 },
  { 'E', 1, 12 },
  { 'F', 4, 2 },
  { 'G', 2, 3 },
  { 'H', 4, 2 },
  { 'I', 1, 9 },
  { 'J', 8, 1 },
  { 'K', 5, 1 },
  { 'L', 1, 4 },
  { 'M', 3, 2 },
  { 'N', 1, 6 },
  { 'O', 1, 8 },
  { 'P', 3, 2 },
  { 'Q', 10, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 4 },
  { 'T', 1, 6 },
  { 'U', 1, 4 },
  { 'V', 4, 2 },
  { 'W', 4, 2 },
  { 'X', 8, 1 },
  { 'Y', 4, 2 },
  { 'Z', 10, 1 },
};

/* french letters */
letter french_letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 3 },
  { 'E', 1, 15 },
  { 'F', 4, 2 },
  { 'G', 2, 2 },
  { 'H', 4, 2 },
  { 'I', 1, 8 },
  { 'J', 8, 1 },
  { 'K', 10, 1 },
  { 'L', 1, 5 },
  { 'M', 2, 3 },
  { 'N', 1, 6 },
  { 'O', 1, 6 },
  { 'P', 3, 2 },
  { 'Q', 8, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 6 },
  { 'T', 1, 6 },
  { 'U', 1, 6 },
  { 'V', 4, 2 },
  { 'W', 10, 1 },
  { 'X', 10, 1 },
  { 'Y', 10, 1 },
  { 'Z', 10, 1 },
};

/* turkish (kelime lik) letters */
letter turkish_kelimelik_letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 12 },
  { 'B', 3, 2 },
  { 'C', 4, 2 },
  { 'D', 3, 2 },
  { 'E', 1, 8 },
  { 'F', 7, 1 },
  { 'G', 5, 1 },
  { 'H', 5, 1 },
  { 'I', 1, 7 },
  { 'J', 10, 1 },
  { 'K', 1, 7 },
  { 'L', 1, 7 },
  { 'M', 2, 4 },
  { 'N', 1, 5 },
  { 'O', 2, 3 },
  { 'P', 5, 1 },
  { 'Q', 0, 0 },
  { 'R', 1, 6 },
  { 'S', 2, 3 },
  { 'T', 1, 5 },
  { 'U', 2, 3 },
  { 'V', 7, 1 },
  { 'W', 0, 0 },
  { 'X', 0, 0 },
  { 'Y', 3, 2 },
  { 'Z', 4, 2 },
  { 'C', 4, 2 }, /* Ç */
  { 'S', 4, 2 }, /* Ş */
  { 'G', 8, 1 }, /* Ğ */
  { 'I', 2, 4 }, /* ı */
  { 'O', 7, 1 }, /* Ö */
  { 'U', 3, 2 }, /* Ü */
};

letter *letters = french_letters;

int value[15][15] = {
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{ TW,  0,  0, DL,  0,  0,  0, DW,  0,  0,  0, DL,  0,  0, TW  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  }};

/* set to 1 for a letter in the board that is a joker */
int is_joker[15][15];

int board[15][15];
int hand[7];
dictionary dict;
dictionary found;
int nb_input_letters;  /* +50 only if nb_input_letters = 7 */

int no50;
int turkish;
int kdist;
int stari, starj;

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

int up(int n)
{
  if (isascii(n)) return toupper(n)-'A'+1;
  if (!turkish) abort();
  //if (n == 105) return 304; /* i => İ */
  if (n == 231 || n == 199) return 27; /* ç => Ç */
  if (n == 351 || n == 350) return 28; /* ş => Ş */
  if (n == 287 || n == 286) return 29; /* ğ => Ğ */
  /* small hack, we don't use I but İ to be ascii-friendly */
  if (n == 305 || n == 304) return 30; /* ı => I */
  if (n == 246 || n == 214) return 31; /* ö => Ö */
  if (n == 252 || n == 220) return 32; /* ü => Ü */
  abort();
}

int my_isalpha(int n)
{
  if (isascii(n)) return isalpha(n);
  if (!turkish) abort();
  return n == 231 || n == 199 ||
         n == 351 || n == 350 ||
         n == 287 || n == 286 ||
         n == 305 || n == 304 ||
         n == 246 || n == 214 ||
         n == 252 || n == 220;
}

int my_islower(int n)
{
  if (isascii(n)) return islower(n);
  if (!turkish) abort();
  return n == 287 ||
         n == 351 ||
         n == 305 ||
         n == 105 ||
         n == 231 ||
         n == 246 ||
         n == 252;
}

void read_value(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  int i, j;
  int c;
  for (i = 0; i < 15; i++)
  for (j = 0; j < 15; j++) {
    do { c = fgetc(f); if (c == EOF) goto bf; } while (isspace(c));
    if (c == '.') c = 0; else {
      if (c == 'd') c = DL;
      else if (c == 't') c = TL;
      else if (c == 'D') c = DW;
      else if (c == 'T') c = TW;
      else if (c == 's') { c = S5; stari = i; starj = j; }
      else goto bf;
    }
    value[i][j] = c;
  }
  fclose(f);
  return;
bf:
  printf("%s: bad file\n", fn); abort();
}

void read_dict(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  while (1) {
    unsigned char n[128];
    int ni[128];
    int pos;
    int i;
    int len;
    if (!fgets((char*)n, 128, f)) break;
    for (i = 0; i < 127 && n[i]; i++) if (isspace(n[i])) { n[i] = 0; break; }
    n[i] = 0;
    if ((len=strlen((char*)n)) == 0) continue;
    for (i = 0, pos = 0; n[pos]; i++, pos+=len) ni[i] = utf8(n+pos, &len);
    ni[i] = 0;
    for (i = 0; ni[i]; i++) ni[i] = up(ni[i]);
    ni[i] = 0;
    add_word(&dict, ni);
  }
  fclose(f);
  return;
}

int fgetc_utf8(FILE *f)
{
  int _;
  unsigned char v[4];
  int c = fgetc(f);
  if (c == EOF) return c;
  if (c < 128) return c;
  v[0] = c;
  if ((c & 0xe0) == 0xc0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf0) == 0xe0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf8) == 0xf0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    v[3] = fgetc(f); if (v[3] == EOF) abort();
    return utf8(v, &_);
  }
  abort();
}

void read_board(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  int i, j;
  int c;
  for (i = 0; i < 15; i++)
  for (j = 0; j < 15; j++) {
    do {
      c = fgetc_utf8(f); if (c == EOF) goto bf;
    } while (isspace(c));
    if (c == '.') c = 0; else {
      if (!my_isalpha(c)) goto bf;
      /* lower letters are joker letters (no point) */
      if (!kelimelik_bot && my_islower(c)) is_joker[i][j] = 1;
      c = up(c);
    }
    board[i][j] = c ? c : -10000000;
  }
  fclose(f);
  return;
bf:
  printf("%s: bad file\n", fn); abort();
}

void read_hand(char *h)
{
  int i, pos, len;
  for (i = 0, pos=0; i < 7; i++, pos+=len) {
    int c = utf8((unsigned char*)h+pos, &len);
    if (c == 0) goto bh;
    if (c != '.') nb_input_letters++;
    if (c == '_') c = 0;                 /* joker */
    else if (c == '.') c = 0x1000000;  /* no letter, mark it as used */
    else { if (!my_isalpha(c)) goto bh; c = up(c); }
    hand[i] = c;
  }
  if (h[pos] && !kelimelik_bot) goto bh;
  return;
bh:
  printf("%s: bad hand\n", h); abort();
}

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
} stack_elem;

int starti, startj, inci, incj;
stack_elem stack[15];
int stacklen;

#if 0
char *getword(void)
{
  int i, j, l;
  static char v[16];
  for (l=0, i=starti, j=startj; l < stacklen; l++, i+=inci, j+=incj)
    v[l] = stack[l].l->l + 'A' - 1 + (board[i][j]!=-10000000 ? 'a'-'A' : 0);
  v[l] = 0;
  return v;
}
#endif

int is_letter(int i, int j)
{
  return i >= 0 && i < 15 && j >= 0 && j < 15 && board[i][j] != -10000000;
}

int line_score(int i, int j, int inci, int incj)
{
  int mulfac = 1;
  int score = 0;
  int nletters = 0;
  int plus_25 = 0;

  while (is_letter(i-inci, j-incj)) i-=inci, j-=incj;

  while (is_letter(i, j)) {
    int letter_mul = 1;
    if (board[i][j] <= 0) switch (value[i][j]) {
    case DW: mulfac *= 2; break;
    case TW: mulfac *= 3; break;
    case DL: letter_mul = 2; break;
    case TL: letter_mul = 3; break;
    case S5: plus_25 = 25; break;
    }
    if (!is_joker[i][j])
      score += letters[abs(board[i][j])].value * letter_mul;
    nletters++;
    i+=inci;
    j+=incj;
  }

  if (nletters == 1) return 0;
  return score * mulfac + plus_25;
}

int score(void)
{
  int s;
  int i, j, l;
  int played_count;
  float sdist;
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (board[i][j] == -10000000) board[i][j] = -stack[l].hl;
  s = line_score(starti, startj, inci, incj);
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj) {
    if (board[i][j] > 0) continue;
    s += line_score(i, j, incj, inci);  /* other direction */
  }
  /* if all letters of (full) hand used, +50 */
  i = 0;
  for (l = 0; l < 7; l++) if (hand[l] & 0x1000000) i++;
  if (i == 7 && nb_input_letters == 7 && no50 == 0) s += 50;
  /* kelimelik adds 30, not 50 */
  if (i == 7 && nb_input_letters == 7 && no50 == 1) s += 30;

  /* in kdist mode, score s becomes s*256 + min distance of new letters to star
   * if nothing on star. If new word is on the star, s becomes s*256+255
   * so that playing on the star is prefered and then play as far from the
   * start as possible for a given number of points
   * sdist is in [0..31], we scale it to [0..255] for more bits
   * (also play more letters in priority, so *8 + nb letters played)
   */
  if (!kdist) goto done;
  if (board[stari][starj] > 0) goto done;
  played_count = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    if (board[i][j] >= 0 || board[i][j] == -10000000) continue;
    played_count++;
  }
  if (board[stari][starj] < 0 && board[stari][starj] != -10000000)
    { s = (s * 256 + 255) * 8 + played_count; goto done; }
  sdist = 31;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    float nd, dx, dy;
    if (board[i][j] >= 0 || board[i][j] == -10000000) continue;
    dx = abs(i - stari);
    dy = abs(j - starj);
    nd = sqrt(dx*dx + dy*dy);
    if (nd < sdist) sdist = nd;
  }
  sdist *= 8;
  s = s * 256 + floor(sdist);
  s *= 8;
  s += played_count;

done:
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (board[i][j] <= 0) board[i][j] = -10000000;
  return s;
}

void check_word(void)
{
  /* to be ok a word must include an already positioned letter or touch one
   * or be put on the center position (can only happen for first word)
   * note that when check_word is called we already checked that putting
   * the letters on the board give existing word (or words)
   */
  int w[4+16];
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (!is_letter(i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<stacklen; l++, i+=inci, j+=incj)
    if (is_letter(i, j) ||             /* word contain letter on board */
        is_letter(i-incj, j-inci) ||   /* word touch letter on board */
        is_letter(i+incj, j+inci) ||   /* word touch letter on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  /* we put score, i, j, inci, at start of word, but! +0x1000000 to avoid 0 */
  w[0] = score() + 0x1000000;
  w[1] = starti + 0x1000000;
  w[2] = startj + 0x1000000;
  w[3] = inci + 0x1000000;
  /* letters coming from joker stored with negative value (see print_word) */
  for (l=0; l<stacklen; l++) w[l+4] = stack[l].l->l * (stack[l].hl ? 1 : -1);
  w[l+4] = 0;
  add_word(&found, w);
//  printf("%s (i %d j %d %s) %d\n", getword(), starti, startj, incj?"horiz":"vert", score());
}

void push(dict_letter *n, int hl, int ni, int nj)
{
  stack[stacklen].l = n;
  stack[stacklen].hl = hl;
  stacklen++;
  /* if n->w we are in an end letter of a word in the dictionary
   * so if on the board there is nothing after current letter,
   * we check that the word is ok here at (i,j)
   */
  if (n->w && (ni==15 || nj==15 || board[ni][nj]==-10000000)) check_word();
}

void pop(void) { stacklen--; }

int other_word_ok(dict_letter *n, int i, int j)
{
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  board[si][sj] = n->l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && board[starti][startj] != -10000000)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && board[endi][endj] != -10000000)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj)
    word[w] = board[i][j];
  word[w] = 0;
  board[si][sj] = -10000000;
//if (w !=1) printf("w %d\n", w);
  return w == 1 || find_word(&dict, word);
}

void go(dict_letter *cur, int i, int j)
{
  /* this function works vertically or horizontally depending on inci/incj */
  /* if in the dictionary the letter at (i,j) matches one in cur,
   * the go to next pos on board and go down in the dictionary tree.
   * the word is checked in push
   */
  int h;

  if (cur == NULL) return;
  if (i == 15 || j == 15) return;

  if (board[i][j] != -10000000) {
    /* must match board[i][j] */
    while (cur) {
      if (cur->l == board[i][j]) {
        push(cur, cur->l, i + inci, j + incj);
        go(cur->child, i + inci, j + incj);
        pop();
        return;
      }
      cur = cur->next;
    }
    return;
  }

  /* must match one of not-yet-used letters */
  for (h = 0; h < 7; h++) {
    dict_letter *n = cur;
    int g;
    if (hand[h] & 0x1000000) continue;  /* letter already used */
    g = hand[h];
    hand[h] |= 0x1000000;
    while (n) {
      /* letter must match or g is joker
       * and (eventual) word in other direction must be ok */
      if ((n->l == g || g == 0) && other_word_ok(n, i, j)) {
        push(n, g, i + inci, j + incj);
        go(n->child, i + inci, j + incj);
        pop();
      }
      n = n->next;
    }
    hand[h] &= ~0x1000000;
  }
}

char *low(int c)
{
  static char r[2];
  if (c <= 26) { r[0] = c + 'a' - 1; r[1] = 0; return r; }
  switch(c) {
  case 27: return "ç";
  case 28: return "ş";
  case 29: return "ğ";
  case 30: return "ı";
  case 31: return "ö";
  case 32: return "ü";
  }
  abort();
}

char *high(int c)
{
  static char r[2];
  if (c <= 26) { r[0] = c + 'A' - 1; r[1] = 0; return r; }
  switch(c) {
  case 27: return "Ç";
  case 28: return "Ş";
  case 29: return "Ğ";
  case 30: return "İ";
  case 31: return "Ö";
  case 32: return "Ü";
  }
  abort();
}

void print_word(int *w)
{
  int i, a, b, inca, incb;
  a = w[1] - 0x1000000;
  b = w[2] - 0x1000000;
  inca = w[3] - 0x1000000;
  incb = 1 - inca;
  if (kelimelik_bot) {
    /* score letters line column is_vertical */
    printf("%d :", w[0] - 0x1000000);
    for (i = 4; w[i]; i++, a+=inca, b+=incb)
      if (board[a][b] == -10000000) printf("%s", low(w[i]));
    printf(" %d %d %d\n", w[1]-0x1000000, w[2]-0x1000000, w[3]-0x1000000);
    return;
  }
  /* print word */
  for (i = 4; w[i]; i++) printf("%s", high(abs(w[i])));
  printf(" lin %d col %d %s ",
         w[1] - 0x1000000, w[2] - 0x1000000,
         w[3] - 0x1000000 ? "verti" : "horiz");
  /* print - for joker letter, UPPER for letter on board (ie. not in hand) */
  for (i = 4; w[i]; i++, a+=inca, b+=incb)
    printf("%s", w[i]<0? "-" : board[a][b] == -10000000? low(w[i]):high(w[i]));
  /* score at field 9 so you can sort by score (is -k gnu sort specific?):
   * ./scrabble <dict> <game> <hand>|sort -k 9nr|less
   */
  printf(" score %d\n",
         w[0] - 0x1000000);
}

void usage(void)
{
  printf("usage: scrabble [options] <dictionary> <board file> <hand>\n");
  printf("options:\n");
  printf("    -english\n");
  printf("        use english letter values\n");
  printf("    -kelimelik\n");
  printf("        to solve kelimelik games (different rules)\n");
  printf("    -value <value board file>\n");
  printf("        mandatory with -kelimelik\n");
  printf("        in the file (15x15, spaces where you want):\n");
  printf("            d/t: double/triple letter\n");
  printf("            D/T: double/triple word\n");
  printf("            s:   +25 at that pos\n");
  printf("            .:   normal pos\n");
  printf("        Copy 'kelimelik.value' and put the 's' stuff in it.\n");
  printf("        The rest (d/t/D/T) does never change.\n");
  printf("    -kbot\n");
  printf("        adapt output for the kelimelik bot\n");
  printf("    -kdist\n");
  printf("        use distance to star in kelimelik to compute score\n");
  exit(1);
}

int main(int n, char **v)
{
#if 0
{
char *t="ğĞşŞıIiİçÇöÖüÜ";
//t = 287 286 351 350 305 73 105 304 231 199 246 214 252 220 
int pos, len, i;
for (i = 0, pos = 0; t[pos]; i++, pos+=len) printf("%d ", utf8((unsigned char*)t+pos,&len));
printf("\n");
return 0;
}
#endif
#if 1
  int i, j;
  int kelimelik = 0;
  char *fdict = NULL;
  char *board = NULL;
  char *hand = NULL;
  char *value = NULL;
  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-english")) { letters = english_letters; continue; }
    if (!strcmp(v[i], "-kbot")) { kelimelik_bot = 1; continue; }
    if (!strcmp(v[i], "-kdist")) { kdist = 1; continue; }
    if (!strcmp(v[i], "-kelimelik")) {
      kelimelik = 1;
      continue;
    }
    if (!strcmp(v[i], "-value"))
      { if (i > n - 2) usage(); value = v[i+1]; i++; continue; }
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (board == NULL) { board = v[i]; continue; }
    if (hand == NULL) { hand = v[i]; continue; }
    usage();
  }
  if (fdict == NULL || board == NULL || hand == NULL) usage();
  if (kelimelik && value == NULL) usage();
  if (kelimelik) {
    no50 = 1;
    turkish = 1;
    letters = turkish_kelimelik_letters;
    read_value(value);
  }
  new_dict(&dict);
  new_dict(&found);
  read_dict(fdict);
  read_board(board);
  read_hand(hand);
  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }
  dump_cb(&found, print_word);
  return 0;
#endif
#if 0
  /* print count of letters (debug stuff) */
  int i;
  n = 0;
  for (i = 0; i < 27; i++) n += letters[i].count;
  printf("%d\n", n); return 0;
#endif
#if 0
  /* output board image (debug stuff for double/triple letter/word)
   * (in array 'value') image format PPM
   */
int col[5][3] = {
  { 20, 110, 20 },
  { 127, 127, 245 },
  { 40, 40, 245 },
  { 215, 127, 127 },
  { 215, 10, 10 }
};

#define WIDTH 20
#define HEIGHT 20
#define BG "30 60 30\n"
  int i, j, I, J;
  printf("P3\n%d %d\n255\n", 15*WIDTH+2, 15*HEIGHT+2);
  printf(BG);
  for (j = 0; j < 15; j++)
  for (J = 0; J < WIDTH; J++)
    printf(BG);
  printf(BG);
  for (i = 0; i < 15; i++)
  for (I = 0; I < HEIGHT; I++) {
    printf(BG);
    for (j = 0; j < 15; j++)
    for (J = 0; J < WIDTH; J++)
      if (I % HEIGHT == 0 || I % HEIGHT == HEIGHT-1 ||
          J % WIDTH == 0 || J % WIDTH == WIDTH-1)
        printf(BG);
      else
        printf("%d %d %d\n", col[value[i][j]][0], col[value[i][j]][1],
                             col[value[i][j]][2]);
    printf(BG);
  }
  printf(BG);
  for (j = 0; j < 15; j++)
  for (J = 0; J < WIDTH; J++)
    printf(BG);
  printf(BG);
  return 0;
#endif
}
